from .colors import Color
from .colors import wrap_iterable_elements_with_color


class Spinner:
    spinners = {
        'classic': (['|', '/', '-', '\\'], 0.1),
        'dots': (['▫ ▫ ▫', '▪ ▫ ▫', '▫ ▪ ▫', '▫ ▫ ▪'], 0.2),
        'equal': (
            ['⁼ ⁼ ⁼ ⁼ ⁼', '= ⁼ ⁼ ⁼ ⁼', '⁼ = ⁼ ⁼ ⁼', '⁼ ⁼ = ⁼ ⁼', '⁼ ⁼ ⁼ = ⁼', '⁼ ⁼ ⁼ ⁼ ='], 0.2),

        'braille': (['⠏', '⠹', '⠼', '⠧'], 0.1),
        'braille_long': (['⡏', '⢹', '⣸', '⣇'], 0.1),
        'braille_crawl': (['⠌', '⠒', '⠡', '⠨'], 0.1),
        'braille_bounce': (['⣶', '⣭', '⠿', '⣭'], 0.1),
        'arc': (['◜', '◝', '◞', '◟'], 0.1),
        'clear_quadrants': (['◴', '◷', '◶', '◵'], 0.1),
    }

    def __iter__(self):
        # Sets internal `_count` to 0 each time we begin a for-based loop:
        self._count = 0
        return self

    def __next__(self):
        # If there are still animations to print, continue printing them:
        if self._count < self._spinner_len:
            spinner_frame = self._spinner[self._count]
            self._count += 1
            return spinner_frame

        # Otherwise, quit:
        raise StopIteration

    def __init__(self, style=None, interval=None, color=None):
        # If style is empty, default to 'classic':
        self._style = style or 'classic'

        while True:
            try:
                # `spinner_tuple` is a tuple containing a list of spinner animations
                # and an interval indicating how often to refresh the animation:
                spinner_tuple = self.spinners[self._style]
                break
            # If the given style doesn't exist, default to 'classic':
            except KeyError:
                self._style = 'classic'

        # `self.interval` is intended to be public:
        self._spinner, self.interval = spinner_tuple

        # If a color is provided, we should color the animation frames:
        if color:
            color_obj = Color(color)
            self._spinner = wrap_iterable_elements_with_color(
                self._spinner, color_obj.color)
        # If user specifies an interval, use that instead:
        if interval:
            self.interval = interval

        # `_spinner_len` is used internally to indicate when the spinner animation
        # should fully reset:
        self._spinner_len = len(self._spinner)



import math
import platform
import shutil
import time

import pytest

from revolution.revolution import Revolution
from revolution.spinner import Spinner


class TestRevolution:
    def setup(self):
        self.revolution_normal = Revolution()

        self.desc = 'Running'
        self.total = 25
        self.style = 'braille'
        self.color = 'cyan'
        self.success = 'Good'
        self.interval = 0.5
        self.revolution_with_params = Revolution(
            desc=self.desc, total=self.total, style=self.style,
            color=self.color, success=self.success, interval=self.interval)

    # Testing __init__
    def test_init_method_variables_for_revolution_normal(self):
        assert self.revolution_normal._count == 0
        assert self.revolution_normal._total == None

        assert self.revolution_normal._desc == ''
        assert self.revolution_normal._desc == ''

        # Test for success:
        assert self.revolution_normal._success_message == ''
        assert self.revolution_normal._success_frame == '\x1b[32m+\x1b[0m'

        # Test for fail:
        assert self.revolution_normal._fail_message == ''
        assert self.revolution_normal._fail_frame == '\x1b[31m-\x1b[0m'

        assert self.revolution_normal._style == 'classic'

        # Test if spinner animation frames are the same:
        spinner = Spinner(self.revolution_normal._style, None, 'blue')
        spinners = [s for s in spinner]
        normal_spinners = [s for s in self.revolution_normal._spinner]
        assert spinners == normal_spinners

        # Test if intervals are the same:
        assert self.revolution_normal._interval == Spinner.spinners[
            self.revolution_normal._style][1]

        assert self.revolution_normal._rate == 0.0

    def test_init__method_variables_for_revolution_with_params(self):
        assert self.revolution_with_params._count == 0
        assert self.revolution_with_params._total == self.total

        # Test description
        assert self.revolution_with_params._desc == self.desc
        assert self.revolution_with_params._message == self.desc

        # Test success/fail
        assert self.revolution_with_params._success_message == self.success
        assert self.revolution_with_params._fail_message == self.desc

        # Test style
        if platform.system() == 'Windows':
            assert self.revolution_with_params._style == 'classic'
        else:
            assert self.revolution_with_params._style == self.style

        spinner = Spinner(self.revolution_with_params._style, None, self.color)
        spinners = [s for s in spinner]
        params_spinners = [s for s in self.revolution_with_params._spinner]
        assert spinners == params_spinners

        # Test the intervals:
        assert self.revolution_with_params._interval == self.interval

        assert self.revolution_with_params._rate == 0.0

    # Testing properties
    def test_success_property(self):
        success_message = 'Complete'

        self.revolution_normal.success = success_message
        assert self.revolution_normal.success == success_message

    def test_fail_property(self):
        fail_message = 'Fail'

        self.revolution_normal.fail = fail_message
        assert self.revolution_normal.fail == fail_message
        assert self.revolution_normal._fail_frame == '\x1b[31m-\x1b[0m'

    # Testing static methods
    def test_make_padding(self):
        num_columns = shutil.get_terminal_size()[0]
        padding = math.floor(num_columns * 0.2)

        assert padding == Revolution.make_padding()

    # Testing public methods
    def test_start_and_stop_threads(self):
        rev = Revolution()

        # Test if main event and thread function properly with `start`:
        rev.start()
        assert not rev._main_event.is_set()
        assert rev._main_thread.is_alive()
        assert rev._main_thread.daemon

        # Test if main event and thread are properly closed with `stop`:
        rev.stop()
        assert rev._main_event.is_set()
        assert not rev._main_thread.is_alive()

    def test_update(self):
        rev1 = Revolution()
        rev2 = Revolution()

        rev1.update()
        assert rev1._count == 1

        rev2.update(2)
        assert rev2._count == 2

    # Testing internal methods
    def test__spin_as_decorator(self):
        @Revolution(desc='As a decorator')
        def description():
            for _ in range(20):
                time.sleep(0.05)

        description()

    def test__spin_with_context_manager(self):
        with Revolution(desc='With a context manager') as rev:
            for _ in range(20):
                time.sleep(0.05)
                rev.update()

        with Revolution(desc='With a context manager and total', total=20, color='red') as rev:
            for _ in range(20):
                time.sleep(0.1)
                rev.update()

    def test__spin_with_for_loop_range(self):
        limit = 20

        expected_total = 0
        for i in range(limit):
            expected_total += i

        total = 0
        for i in Revolution(range(20)):
            total += i

        assert total == expected_total

    def test__spin_with_for_loop_list(self):
        numbers = list(range(20))

        expected_total = 0
        for i in numbers:
            expected_total += i

        total = 0
        for i in Revolution(numbers):
            total += i

        assert total == expected_total

    def test__spin_manual_use_with_success(self):
        rev = Revolution(desc='Manual use (success)',
                         style='braille', safe=False)
        rev.start()
        time.sleep(2)

        rev.success = 'Done!'

        rev.stop()

    def test__spin_manual_use_with_fail(self):
        rev = Revolution(desc='Manual use (fail)')
        rev.start()
        time.sleep(2)

        rev.fail = 'Unfortunate'

    def test__make_statement_for_empty_statement(self):
        statement = self.revolution_normal._make_statement()
        empty_statement = '{} {}'

        assert empty_statement in statement

    def test__make_statement_for_statement_with_total(self):
        statement = self.revolution_with_params._make_statement()
        statement_with_total = '{} {} « {}/{} » [{}]'

        assert statement_with_total in statement



import pytest

from revolution.colors import Color
from revolution.colors import wrap_text_with_color, wrap_iterable_elements_with_color


class TestColor:
    def setup(self):
        self.color = 'red'
        self.color_obj = Color(self.color)

    # Test __init__
    def test_init_correct(self):
        assert self.color_obj._color == (Color.colors[self.color], Color.reset)

    # Test hidden methods
    def test__verify_color(self):
        assert self.color_obj._verify_color('blue') == (
            Color.colors['blue'], Color.reset)

    # Test properties
    def test_color_property(self):
        # getter
        assert self.color_obj.color == (Color.colors[self.color], Color.reset)
        # setter
        self.color_obj.color = 'blue'
        assert self.color_obj.color == (Color.colors['blue'], Color.reset)


def test_wrap_text_with_color():
    red = 'red'
    color = Color(red)

    text = 'Test'

    new_text = wrap_text_with_color(text, color.color)
    assert new_text == f'{Color.colors[red]}{text}{Color.reset}'


def test_wrap_iterable_elements_with_color():
    red = 'red'
    color = Color(red)

    spinners = ['-', '\\', '|', '/']
    new_spinners = wrap_iterable_elements_with_color(spinners, color.color)
    colored_spinners = ['\x1b[31m-\x1b[0m', '\x1b[31m\\\x1b[0m',
                        '\x1b[31m|\x1b[0m', '\x1b[31m/\x1b[0m']
    assert new_spinners == colored_spinners



import platform
import subprocess


class Color:
    reset = '\x1b[0m'

    colors = {
        'black': '\x1b[30m',
        'red': '\x1b[31m',
        'green': '\x1b[32m',
        'yellow': '\x1b[33m',
        'blue': '\x1b[34m',
        'violet': '\x1b[35m',
        'cyan': '\x1b[36m',
        'white': '\x1b[37m'
    }

    def __new__(cls, color):
        """
        Overrides default __new__ method to enable color support on Windows machines.
        """

        # Enables color support for Windows machines:
        if platform.system() == 'Windows':
            subprocess.call('color', shell=True)

        instance = super(Color, cls).__new__(cls)
        return instance

    def __init__(self, color):
        self._color = self._verify_color(color)

    @property
    def color(self) -> tuple:
        return self._color

    @color.setter
    def color(self, color):
        self._color = self._verify_color(color)

    def _verify_color(self, color) -> tuple:
        # Clean `color` input first:
        cleaned_color = color.lower().strip()

        if isinstance(color, str):
            # Search for color in `colors`:
            if cleaned_color in Color.colors:
                return (Color.colors[cleaned_color], Color.reset)
            else:
                raise KeyError(
                    f'Color not found. Valid colors: {list(Color.colors)}')
        else:
            raise TypeError(
                f'color parameter of type {type(color)} must be str')


def wrap_text_with_color(text, color: tuple) -> str:
    ascii_color, ascii_reset = color
    text_fmt = '{}{}{}'

    return text_fmt.format(ascii_color, text, ascii_reset)


def wrap_iterable_elements_with_color(iterable, color: tuple) -> list:
    ascii_color, ascii_reset = color
    text_fmt = '{}{}{}'

    colored_iterable = [
        text_fmt.format(ascii_color, element, ascii_reset) for element in iterable]

    return colored_iterable



import math
import platform
import shutil
import sys
import threading
import time
import traceback
import types

from .spinner import Spinner
from .colors import Color
from .colors import wrap_text_with_color


class Revolution:
    def __call__(self, *args, **kwargs):
        """
        Allows `Revolution` to be called as a function decorator.
        """

        # If the decorator doesn't have any arguments:
        if hasattr(self, '_func'):
            self.start()
            result = self._func(*args, **kwargs)
            self._main_event.set()
            while not self._spin_event.is_set():
                pass
            return result
        # ...otherwise, the function is in *args:
        else:
            func = args[0]
            if isinstance(func, types.FunctionType):
                def wrapper(*margs, **mkwargs):
                    self.start()

                    result = func(*margs, **mkwargs)

                    if self._total:
                        self._count += 1
                        if self._total == self._count:
                            while not self._spin_event.is_set():
                                pass
                            return result
                    else:
                        self._main_event.set()
                        while not self._spin_event.is_set():
                            pass
                        return result
                return wrapper

    def __enter__(self):
        """
        Entry point for with statements. Used in conjunction with __exit__.
        """

        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Exit point for with statements. Used in conjunction with __enter__.
        """

        self._main_event.set()
        if (exc_type, exc_value, exc_traceback) == (None, None, None):
            while not self._spin_event.is_set():
                pass
            return
        traceback.print_exception(exc_type, exc_value, exc_traceback)

    def __iter__(self):
        """
        Allows for loops to be used with Revolution objects. Used in conjunction with
        __next__.
        """

        self.start()
        return self

    def __next__(self):
        """
        Used in conjunction with with __iter__.
        """

        # If the for loop was given an iterable such as a list:
        if hasattr(self, '_iter'):
            if self._count < self._stop:
                return_value = self._iter[self._count]
                self._count += self._step
                return return_value
            raise StopIteration
        # otherwise, it's likely a range object and so:
        else:
            if self._count < self._stop:
                return_value = self._count
                self._count += self._step
                return return_value
            raise StopIteration

    def __init__(self, func=None, desc='', total=None, style='',
                 color='blue', success=None, safe=True, interval=None):
        """
        Initializes a Revolution object.

        Parameters
        ==========
        func : list or range, optional
            If this is a list or range object, it will iterate over each of the elements
            and return them one by one.

            This parameter is necessary to be able to use a Revolution object as a function
            decorator.

        desc : str, optional
            A string to use in place of the text that displays beside the spinner.

        total : int, optional
            An integer that indicates the total number of expected iterations.

            It's recommended that you do not include this when using a Revolution object
            as a function decorator.

        style : str, optional
            A string that indicates which spinner style to use. If style is None or if it
            doesn't exist, the classic style will be used.

            Available options can be viewed by running `revolution --example` or
            `revolution -e` in your console.

        color : str, optional
            A string that indicates which color should be used for the spinner. If a color
            is not provided, the color will default to 'blue'.

            Available options:
                * 'black'
                * 'red'
                * 'green'
                * 'yellow'
                * 'blue'
                * 'violet'
                * 'cyan'
                * 'white'

        success : str, optional
            A string that will be displayed beside the spinner when the spinner animation
            stops.

        safe : bool, optional
            If True (default), spinners on Windows machines will always use the
            'classic' style (even if a different style is provided). 

            If you are using a certain spinner style and are unsure as to how it will 
            appear on Windows machines, it is recommended that you leave `safe` set to
            its default value.

        interval : float, optional
            A float value that is used to indicate the refresh rate of the entire spinner.
        """

        if func:
            # If `func` is provided a range object or list:
            if isinstance(func, range) or isinstance(func, list):
                try:
                    self._count = func.start
                    self._stop = func.stop
                    self._step = func.step

                    self._total = func.end
                except AttributeError:
                    self._count = 0
                    self._stop = len(func)
                    self._step = 1
                    self._iter = func

                    self._total = self._stop
            # ...otherwise, it must be a FunctionType:
            else:
                self._func = func
                self._count = 0
                self._total = total
        else:
            self._count = 0
            self._total = total

        self._desc = desc
        self._message = desc

        # For success messages:
        self._success_message = success or desc
        green = Color('green')
        self._success_frame = wrap_text_with_color('+', green.color)
        self._frame = self._success_frame

        # For fail messages:
        self._fail_message = desc
        red = Color('red')
        self._fail_frame = wrap_text_with_color('-', red.color)

        # Determines which spinner style should be used based on the OS:
        if safe:
            if platform.system() == 'Windows':
                self._style = 'classic'
            else:
                self._style = style or 'classic'
        else:
            self._style = style or 'classic'

        self._spinner = Spinner(self._style, interval, color)
        self._interval = self._spinner.interval

        self._rate = 0.0

    @property
    def success(self):
        return self._success_message

    @success.setter
    def success(self, text):
        self._success_message = text
        self._message = self._success_message

        try:
            self.stop()
        except AttributeError:
            pass

    @property
    def fail(self):
        return self._fail_message

    @fail.setter
    def fail(self, text):
        self._fail_message = text
        self._message = self._fail_message
        self._frame = self._fail_frame

        try:
            self.stop()
        except AttributeError:
            pass

    @staticmethod
    def make_padding() -> int:
        """
        Returns number of columns to use for spacing.
        """
        num_columns = shutil.get_terminal_size()[0]
        padding = math.floor(num_columns * 0.2)
        return padding

    def start(self):
        """
        Starts the method responsible for animating spinner frames to the console.

        If this method is used manually and externally, the user is responsible for
        calling the `stop` method as well (or `success` or `fail`).
        """

        self._main_event = threading.Event()
        # self._main_thread targets the function responsible for animating frames:
        self._main_thread = threading.Thread(target=self._spin)

        # Set self._main_thread as daemon so this internal thread will quit if the
        # main thread quits:
        self._main_thread.setDaemon(True)
        self._main_thread.start()

    def stop(self):
        """
        Stops the method responsible for animating spinner frames to the console.

        If a user manually calls a Revolution object's `start` method, then they
        are responsible for calling that object's `stop` method as well.
        """

        if not self._main_event.is_set():
            self._main_event.set()
            self._main_thread.join()

    def update(self, step=1):
        """
        Updates the `_count` member variable.

        Parameters
        ==========
        step : int, optional
            Indicates by how much the _count member variable should be incremented.

        Example
        =======
        >>> from revolution import Revolution
        >>> rev = Revolution()
        >>> rev.start()
        >>> rev.update()
        >>> rev.update(2)
        >>> rev.stop()
        """
        self._count += step

    def _spin(self):
        """
        Prints a spinner animation frame to the console.
        """

        # For preventing premature ejection:
        self._spin_event = threading.Event()

        # Enable thread for measuring rate:
        self._rate_event = threading.Event()
        rate_thread = threading.Thread(target=self._update_rate)
        rate_thread.setDaemon(True)
        rate_thread.start()

        self._statement = self._make_statement()

        while True:
            for frame in self._spinner:
                print('\r', end='')
                sys.stdout.write(self._statement.format(
                    frame, self._desc, self._count, self._total, self._rate))

                if self._main_event.is_set() or self._count == self._total:
                    print('\r', end='')
                    sys.stdout.write(self._statement.format(
                        self._frame, self._message, self._count, self._total, self._rate) + '\n')

                    self._rate_event.set()
                    self._spin_event.set()
                    return

                time.sleep(self._interval)

    def _make_statement(self):
        """
        Returns a string containing the format to display based on whether a Revolution
        instance's `total` parameter is provided a value.
        """

        if self._total:
            statement = ' {} {} « {}/{} » [{}]' + (' ' * self.make_padding())
        else:
            statement = ' {} {}' + (' ' * self.make_padding())
        return statement

    def _update_rate(self):
        """
        Calculates the number of iterations per second by dividing the current `_count`
        by the running time of the spinner.
        """

        if self._total:
            start = time.perf_counter()
            while not self._rate_event.is_set():
                self._rate = f'{round(self._count / (time.perf_counter() - start), 4)} it/s'
                time.sleep(0.25)



import pytest

from revolution.spinner import Spinner


def test_style_becomes_style_if_style_is_provided():
    style = 'equal'
    spinner = Spinner(style=style)
    assert spinner._style == style


def test_style_becomes_classic_if_style_is_not_provided():
    spinner = Spinner()  # `style` not specified
    assert spinner._style == 'classic'


def test_spinner_becomes_classic_if_style_does_not_exist():
    nonexistent_style = 'shimmalummadingdong'
    spinner = Spinner(style=nonexistent_style)
    assert spinner._spinner == Spinner.spinners['classic'][0]


def test_spinner_interval_becomes_interval_if_interval_is_provided():
    my_spinner_interval = 0.75

    spinner = Spinner(interval=my_spinner_interval)
    assert spinner.interval == my_spinner_interval


def test_spinner_interval_becomes_default_if_interval_is_not_provided():
    classic_spinner_interval = Spinner.spinners['classic'][1]

    classic_spinner = Spinner(style='classic')
    assert classic_spinner.interval == classic_spinner_interval


def test_spinner_len_equals_length_of_spinner_in_spinners():
    classic_spinner_len = len(Spinner.spinners['classic'][0])

    classic_spinner = Spinner(style='classic')
    assert classic_spinner._spinner_len == classic_spinner_len


def test_spinner_iterates_over_animations_when_used_as_iterator():
    num_classic_spinner_animations = len(Spinner.spinners['classic'][0])

    classic_spinner = Spinner(style='classic')

    count = 0
    for _ in classic_spinner:
        count += 1
    assert count == num_classic_spinner_animations
